/**
   From the paper "A Java Fork-Join Framework" by Doug Lea.

   This has been modified to use Java 7's Fork-Join Framework.
*/
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveAction;

@SuppressWarnings("serial")
class Fib extends RecursiveAction // which extends java.util.concurrent.ForkJoinTask<V>
{
   public static final int THRESHOLD = 15;  // BE SURE TO EXPERIMENT WITH THIS!
   volatile long number; // arg/result

   Fib(long number)
   {
      this.number = number;
   }


   long getAnswer()
   {
      if (!isDone())  // from java.util.concurrent.ForkJoinTask<V>
         throw new IllegalStateException();
      return number;
   }


   @Override
   protected void compute()
   {
      long n = number;
      if (n <= THRESHOLD) // granularity ctl
      {
         number = seqFib(n);
      }
      else
      {
         Fib f1 = new Fib(n - 1);
         Fib f2 = new Fib(n - 2);
       //coInvoke(f1, f2);        // dosn't compile
       //invokeAll(f1, f2);       // this works
         f1.fork();    // run f1 asynchronously
         f2.compute(); // run f2 synchronously
         f1.join();    // wait for f1 to complete
         number = f1.number + f2.number;
      }
   }


   static long seqFib(long n)
   {
      if (n <= 1)
         return n;
      else
         return seqFib(n-1) + seqFib(n-2);
   }
}


class Fib_Thread extends Thread
{
   //public static final int THRESHOLD = 30;

   public volatile long number;  // why should we use volatile?
   public volatile long result;

   public Fib_Thread(long number)
   {
      this.number = number;
      this.result = -1;
   }


   @Override
   public void run()
   {
      if (this.number < Fib.THRESHOLD)
      {
         this.result = Fib.seqFib(number);
      }
      else
      {
         Fib_Thread f1 = new Fib_Thread(number - 1);
         Fib_Thread f2 = new Fib_Thread(number - 2);
         f1.start();   // run f1 asynchronously
         f2.run();     // run f2 synchronously
         try{
         f1.join();    // wait for f1 to complete
         }catch(InterruptedException e){}
         result = f1.result + f2.result;
      }
   }
}



public class Fib_Java7
{
   public static void main(String[] args)
   {
      if (0 == args.length)
      {
         System.err.println("Usage: java Fib_Java7 <integer>");
         System.exit(-1);
      }
      int n = Integer.parseInt(args[0]);

      long startTime, stopTime;

      // Run the single-threaded sequential version.
      startTime = System.currentTimeMillis();
      long result = Fib.seqFib(n);      //  synchronous function call
      stopTime = System.currentTimeMillis();
      System.out.println("Sequential time: " + (stopTime - startTime));
      System.out.println("Sequential answer: " + result);

      // Run the fork-join version.
      ForkJoinPool forkJoinPool = new ForkJoinPool();
      Fib f = new Fib(n);
      startTime = System.currentTimeMillis();
      forkJoinPool.invoke(f);        //  synchronous task
      result = f.getAnswer();
      stopTime = System.currentTimeMillis();
      System.out.println(" Fork-Join time: " + (stopTime - startTime));
      System.out.println(" Fork-Join answer: " + result);

      // Run the multi-threaded version.
      Fib_Thread t = new Fib_Thread(n);
      startTime = System.currentTimeMillis();
      t.start();                       //  asynchronous thread
      try { t.join(); }catch(InterruptedException e){}
      result = t.result;
      stopTime = System.currentTimeMillis();
      System.out.println("Multi-threaded time: " + (stopTime - startTime));
      System.out.println("Multi-threaded answer: " + result);
   }
}